from __future__ import division
from scipy.io import wavfile
from IPython.core.display import HTML
from IPython.display import display
import numpy as np
import matplotlib.pyplot as plt
from heapq import heappush, heappop, heapify
from collections import defaultdict

# this is a wrapper that take a filename and publish an html <audio> tag to listen to it
def wavPlayer(filepath):
    src = """
    <head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8">
    <title>Simple Test</title>
    </head>

    <body>
    <audio controls="controls" style="width:600px" >
      <source src="files/%s" type="audio/wav" />
      Your browser does not support the audio element.
    </audio>
    </body>
    """%(filepath)
    display(HTML(src))

def playMusic(title, rate, music):
    # write the file on disk, and show in in a Html 5 audio player
    music_wav = music * 2**13
    wavfile.write(title + '.wav', rate, music_wav.astype(np.int16))
    wavPlayer(title + '.wav')

def generate_quantized_song(num_of_bits):
    # Code to create song
    # C:40, D:42, E:44, F:45, G:47, A:49, B:51, Higher C:52
    notes = [47, 47, 49, 49, 47, 47, 44, 44, 47, 47, 44, 44, 42, 42, 42, 0,
             47, 47, 49, 49, 47, 47, 44, 44, 47, 44, 42, 44, 40, 40, 40, 0]
    note_duration = 0.5 # per key, in sec
    fs = 1200
    music_duration = note_duration * len(notes)
    time = np.linspace(0, music_duration, num=fs*music_duration) # indices for 0-16 secs spaced apart by 1/44100
    song = np.zeros(len(time))

    sinwave = lambda f,t : np.sin(2*np.pi*f*t) # generate a sinusoid of frequency f Hz over time t
    key2freq = lambda n : 2 ** ((n-49)/12) * 440 # convert a piano key n to a frequency in Hz

    idx_note = 0
    for note in notes:
        if note: # if note == 0, skip.
            freq = key2freq(note) # frequency of each note
            song[idx_note*fs*note_duration : (idx_note+1)*fs*note_duration-1] = \
                    sinwave( freq,
                            time[idx_note*fs*note_duration : (idx_note+1)*fs*note_duration-1 ]
                            )
        idx_note += 1
    quantized_song = quantizer(num_of_bits, song)
    return time, quantized_song

def quantizer(num_of_bits, original_signal): # kangwook's fast quantizer
    size_of_two_levels = (2**num_of_bits-1) # (L - 1)
    quantized_signal = (np.round((original_signal * size_of_two_levels + 1)/2)*2-1) / size_of_two_levels
    return quantized_signal


class Tree:
    def __init__(self, left, right, value):
        self.left = left
        self.right = right
        self.symbol = value
    @staticmethod
    def node(value):
        return Tree(None, None, value)
    @staticmethod
    def tree(left, right):
        return Tree(left, right, None)
    def mapping(self, prefix=''):
        if self.symbol:
            return {self.symbol: prefix}
        left = self.left.mapping(prefix + '0')
        right = self.right.mapping(prefix + '1')
        return dict(left, **right)

def HuffEncode(freq_dict):
    """Return a dictionary (letters2huff) which maps keys from the input dictionary freq_dict
       to bitstrings using a Huffman code based on the frequencies of each key"""
    heap = []
    for symbol, freq in freq_dict.items():
        heappush(heap, (freq, Tree.node(symbol)))
    while len(heap) > 1:
        left_f, left = heappop(heap)
        right_f, right = heappop(heap)
        heappush(heap, (left_f + right_f, Tree.tree(left, right)))
    root_f, root = heappop(heap)

    return root.mapping()

def reverse_dict(original_dict):
    # create a new dict with the keys of original_dict as values and the values of original_dict as keys
    new_dict = {}
    for key in original_dict:
        new_dict[original_dict[key]] = key
    return new_dict

def encode_song(song, huff):
    encoded = ''
    for value in song:
        encoded += huff[huff.keys()[np.argmin(abs([huff.keys()] - value))]]
    return encoded

def decode_song(song, huff):
    dhuff = reverse_dict(huff)
    original = np.array([])
    current = ''
    for i in song:
        current += i
        if current in dhuff:
            original = np.r_[original, dhuff[current]]
            current = ''
    return original